/* -*- mode: c++; c-basic-offset: 2; -*- */

#ifndef GENERAL__THREAD_HH
#define GENERAL__THREAD_HH

#include <limits.h>
#include <signal.h>

#include <stdexcept>

#include "general/mutexlock.hh"

namespace General
{
  class Thread
  {
  public:
    class cancellation
      : public std::runtime_error
    {
    public:
      cancellation( const std::string& Header,
		   const char* File, const int Line );
    private:
      static std::string format( const std::string& Header,
				 const char* File, const int Line );
    };

    class deadlock
      : public std::logic_error
    {
    public:
      deadlock( const std::string& Header,
		const char* File, const int Line );
    private:
      static std::string format( const std::string& Header,
				 const char* File, const int Line );
    };

    class invalid_argument
      : public std::invalid_argument
    {
    public:
      invalid_argument( const std::string& Header,
			const char* File, const int Line );
    private:
      static std::string format( const std::string& Header,
				 const char* File, const int Line );
    };

    class range_error
      : public std::range_error
    {
    public:
      range_error( const std::string& Header,
		   const char* File, const int Line );
    private:
      static std::string format( const std::string& Header,
				 const char* File, const int Line );
    };

    typedef enum {
      CANCEL_ASYNCHRONOUS,
      CANCEL_DEFERRED,
      CANCEL_EXCEPTION
    } cancel_type;

    typedef pthread_t thread_type;
    typedef void (*cleanup_function_type)( void* );

    Thread( );

    virtual ~Thread( );

    virtual void Cancel( );

    //-------------------------------------------------------------------
    /// \brief Deliver pending cancellation requests to the calling thread.
    ///
    /// \param[in] Header
    ///     Description of the method from which the request was made
    /// \param[in] File
    ///     Filename from where the request was made.
    /// \param[in] Line
    ///     Line number in Filename from where the request was made.
    //-------------------------------------------------------------------
    void  CancellationCheck( const std::string& Header,
			     const char* File, const int Line ) const;

    void CancellationEnable( bool Value );

    void CancellationType( cancel_type Type );

    void Detach( ) const;

    virtual void Join( );

    bool IsAlive( ) const;

    bool IsParentThread( ) const;

    int Kill( int Signal ) const;

    //-------------------------------------------------------------------
    /// \brief Retrieve key for parent thread
    //-------------------------------------------------------------------
    const thread_type ParentThread( ) const;

    virtual int Spawn( );

    static Thread& Self( );

    static Thread& Self( thread_type Key );
    
    static void  SelfCancellationCheck( const std::string& Header,
					const char* File, const int Line );

    static size_t StackSizeDefault( );

    static void StackSizeDefault( size_t StackSize );

    bool operator==( const Thread& Source ) const;

  protected:
    typedef void *(*start_function_type)( void* );

    //-------------------------------------------------------------------
    /// \brief Job to be acomplished by the thread
    ///
    //-------------------------------------------------------------------
    virtual void action( ) = 0;

    //-------------------------------------------------------------------
    /// \brief Spawn a new thread
    ///
    /// \param[in] StartFunction
    ///     The routine for the new thread to start.
    //-------------------------------------------------------------------
    int spawn( start_function_type StartFunction
	       = (start_function_type)start_routine );

    const thread_type threadId( ) const;

  private:
    const thread_type			m_parent_tid;
    mutable thread_type			m_tid;
    mutable MutexLock::lock_type	m_key;
    //-------------------------------------------------------------------
    /// \brief Canel thread using exceptions
    //-------------------------------------------------------------------
    mutable bool			m_cancel_via_exception;
    //-------------------------------------------------------------------
    /// \brief Enable or disable the cancellability of a thread
    //-------------------------------------------------------------------
    mutable bool			m_cancel_state;
    //-------------------------------------------------------------------
    /// \brief Conditional state of thread being cancelled.
    //-------------------------------------------------------------------
    mutable bool			m_cancel_thread;

    static size_t			m_stack_size_default;

    static Thread* start_routine( Thread* ThreadSource );

  };

  inline bool Thread::
  IsAlive( ) const
  {
    return ( Kill( 0 ) == 0 );
  }

  inline const Thread::thread_type Thread::
  ParentThread( ) const
  {
    return m_parent_tid;
  }

  inline void Thread::
  SelfCancellationCheck( const std::string& Header,
			 const char* File, const int Line )
  {
    try
    {
         Thread& self( Self( ) );

	 self.CancellationCheck( Header, File, Line );
    }
    catch( const Thread::cancellation& Exception )
    {
      //-----------------------------------------------------------------
      // Rethow this error since it means the thread is to be cancelled.
      //-----------------------------------------------------------------
      throw;
    }
    catch( ... )
    {
      //-----------------------------------------------------------------
      // Everything else is caught and quietly ignored.
      //-----------------------------------------------------------------
    }
  }

  inline size_t Thread::
  StackSizeDefault( )
  {
    return m_stack_size_default;
  }

  inline void Thread::
  StackSizeDefault( size_t StackSize )
  {
    m_stack_size_default = StackSize;
  }

  inline const Thread::thread_type Thread::
  threadId( ) const
  {
    return m_tid;
  }
}

#endif /* GENERAL__THREAD_HH */
